from .interface import EngineInterface, MiddlewareInterface
from typing import Any, Callable, List, Optional, Union

from logging import getLogger
logger = getLogger(__name__)

class Middleware(MiddlewareInterface):
    """
    中间件基类
    注意 next_middleware 参数是可选的,如果未提供,可能有两种模式

    1.如果process返回值为None(只有next_middleware为None返回值才为None)
    则返回数据将在Pipeline里返回,否则继续执行Pipeline的下一个中间件。

    2.中间件的process返回值为process_logic处理后的值
    """
    def __init__(self, process_logic: Callable[[Any], Any],
                 next_middleware: Optional[List[MiddlewareInterface]] = []):
        self._process_logic = process_logic
        self._next_middleware = next_middleware
        self._data = None

    def process(self, data: Any) -> Any:
        """处理数据并返回处理后的结果,支持自动调用下一个中间件
        (如果next_middleware为None则返回None)"""
        try:
            data = self._process_logic(data)
            self._data = data
            if self._next_middleware:
                for middleware in self._next_middleware:
                    self._data = middleware.process(self._data)
            return None if self._next_middleware is None else data
        except Exception as e:
            raise e

    def get_next(self) -> Optional[List[MiddlewareInterface]]:
        return self._next_middleware

class Engine(EngineInterface):
    """
    引擎基类

    注意在继承该类时需要注意使用的时特有的格式，方便上下文传输

    下面示例给出继承时的基本操作
    ```python
    class MyEngine(Engine):
        def __init__(is_stream: bool = False):
            super().__init__(is_stream)
        
        def execute(date):
            # you need to do something
            pass
    ```

    高级操作
    ```python
    class MyEngine(Engine):
        def __init__(is_stream: bool = False):
            super().__init__(is_stream)
        
        def prepare_input(data):
            return super().prepare_input(data)
        
        def process_output(data):
            return super().process_output(data)

        def execute(date):
            # you need to do something
            pass
    ```

    """
    def __init__(self) -> None:
        self._metadata = None

    def prepare_input(self, data: Any) -> Any:
        self._metadata = data.get('__metadata__', None)
        return data['__data__']

    def process_output(self, data: Any) -> Any:
        return {'__data__': data, '__metadata__': self._metadata}

    def execute(self, data: Any) -> Any:
        raise NotImplementedError("execute method not implemented")

    def to_middleware(self, next_middleware: Optional[List[MiddlewareInterface]] = []) -> MiddlewareInterface:
        return Middleware(lambda data:
                          self.process_output(
                              self.execute(
                                  self.prepare_input(data))), next_middleware)

    def to_prepare_input_middleware(self, next_middleware: Optional[List[MiddlewareInterface]] = []) -> MiddlewareInterface:
        return Middleware(self.prepare_input, next_middleware)
    def to_process_output_middleware(self, next_middleware: Optional[List[MiddlewareInterface]] = []) -> MiddlewareInterface:
        return Middleware(self.process_output, next_middleware)
    def to_execute_middleware(self, next_middleware: Optional[List[MiddlewareInterface]] = []) -> MiddlewareInterface:
        return Middleware(self.execute, next_middleware)

class ExecutePipeline:
    """
    执行流 用于按顺序执行中间件
    
    示例:
    ```python
    pipeline = ExecutePipeline(middleware1, middleware2)
    pipeline.add(middleware3)
    result = pipeline.execute(data)
    ```
    执行顺序 data -> middleware1 -> middleware2 -> middleware3 -> result

    ```python
    pipeline.reorder([0,2,1])
    result = pipeline.execute(data)
    ```
    执行顺序 data -> middleware1 -> middleware3 -> middleware2 -> result

    ```python
    pipeline.reorder([0,1])
    result = pipeline.execute(data)
    ```
    执行顺序 data -> middleware1 -> middleware2 -> result
    """
    def __init__(self, *components: Union[MiddlewareInterface, EngineInterface]):
        self.middlewares: List[MiddlewareInterface] = []
        for component in components:
            if isinstance(component, EngineInterface):
                self.middlewares.append(component.to_middleware())
            elif isinstance(component, MiddlewareInterface):
                self.middlewares.append(component)
            else:
                raise TypeError(f"Unsupported type {type(component)}")

    def add(self, middleware: MiddlewareInterface):
        """添加中间件到执行流里"""
        self.middlewares.append(middleware)

    def reorder(self, new_order: List[int]):
        """重新排序中间件,按照new_order指定的索引顺序排列,注意可以实现删除功能"""
        self.middlewares = [self.middlewares[i] for i in new_order]

    def execute(self, data):
        """执行中间件

        注意中间件的next不参与主执行流

        注意只有中间件的next为None,返回值才不会被输入到执行流中,即停止执行后续内容

        TODO: 1. 添加中间件执行失败时的处理机制
        """
        try:
            for middleware in self.middlewares:
                if data is None:
                    break
                data = middleware.process(data)
            return data
        except ValueError:
            return None

    def execute_engines(self, data: Any, metadata = None) -> Any:
        """
        执行引擎
        """
        res = self.execute({'__data__': data, '__metadata__': metadata})
        if not res:
            return None
        else:
            return res['__data__']
